"""
Created on 18.01.2023

@author: wf
"""

import sys
from dataclasses import dataclass
from typing import List

from meta.metamodel import Context
from meta.mw import SMWAccess

from tests.basemwtest import BaseMediawikiTest


class ContextContext:

    def __init__(self, smwAccess: SMWAccess, context: Context, debug: bool = False):
        self.smwAccess = smwAccess
        self.context = context
        self.debug = debug

    debug: bool = False


class BaseSemanticMediawikiTest(BaseMediawikiTest):
    """
    special semantic mediawiki tests
    """

    def setUp(self, debug=False, profile=True):
        BaseMediawikiTest.setUp(self, debug=debug, profile=profile)

    def check_contexts(self, wikiId: str, ignoreExceptions: bool = True):
        """
        check the mediawiki contexts  for the given wikiId

        Args:
            wikiId(str): the wikiId to check
            ignoreExceptions(bool): if True only print out a warning other wise raise exception
        """
        mw_contexts = {}
        smwAccess=None
        try:
            smwAccess = SMWAccess(wikiId, debug=self.debug)
            mw_contexts = smwAccess.getMwContexts()
            if self.debug:
                for name, mw_context in mw_contexts.items():
                    print(f"{name}: {mw_context.sidif_url()}")
                print(f"found {len(mw_contexts)} Contexts")
        except Exception as ex:
            print(
                f"warning Mediawiki context for {wikiId} fetching failed due to {str(ex)}",
                file=sys.stderr,
            )
            if not ignoreExceptions:
                raise ex
        return smwAccess, mw_contexts

    def getContextContext(self, wikiId: str = "wiki", context_name: str = "MetaModel"):
        """
        get the default meta model context
        """
        smwAccess, mw_contexts = self.check_contexts(wikiId)
        mw_context = mw_contexts[context_name]
        context, error, _errMsg = Context.fromWikiContext(mw_context, debug=self.debug)
        self.assertIsNone(error)
        cc = ContextContext(smwAccess, context)
        return cc



"""
Created on 2022-11-23

@author: wf
"""

import json

from wikibot3rd.wikiuser import WikiUser

from meta.metamodel import Context
from tests.basesmwtest import BaseSemanticMediawikiTest


class TestMediawiki(BaseSemanticMediawikiTest):
    """
    test reading SiDIF from Mediawiki pages
    """

    def setUp(self, debug=False, profile=True):
        """
        setUp
        """
        BaseSemanticMediawikiTest.setUp(self, debug=debug, profile=profile)
        for wikiId in ["wiki", "royals", "cr"]:
            self.getSMWAccess(wikiId, save=True)

    def test_contexts(self):
        """
        test getting the list of mediawiki contexts
        """
        _smwAccess, mw_contexts = self.check_contexts("wiki")
        self.assertTrue(len(mw_contexts) >= 15)

    def test_conceptProperty(self):
        """
        test the conceptProperty handling
        """
        debug=self.debug
        cc = self.getContextContext()
        topic = cc.context.topics["Property"]
        self.assertIsNotNone(topic.conceptProperty)
        if debug:
            print(topic.conceptProperty)
        self.assertEqual("name", topic.conceptProperty.name)

    def test_sortProperties(self):
        """
        test the access to the list of sort properties
        """
        cc = self.getContextContext()
        topic = cc.context.topics["Property"]
        sorted_props = topic.sortProperties()
        self.assertEqual(1, len(sorted_props))
        sort_prop = sorted_props[0]
        self.assertEqual("name", sort_prop.name)

    def test_propertiesByIndex(self):
        """
        test the properties by Index access
        """
        debug=self.debug
        cc = self.getContextContext()
        topic = cc.context.topics["Topic"]
        props = topic.propertiesByIndex()
        if debug:
            for prop in props:
                print(f"{getattr(prop,'index','?')}:{prop.name}")
        self.assertEqual(13, len(props))
        self.assertTrue("pluralName", props[1].name)

    def testAskQuery4Topic(self):
        """
        test creating the ask query for a topic
        """
        debug = self.debug
        #debug = True
        cc = self.getContextContext()
        topic = cc.context.topics["Property"]
        ask_query = topic.askQuery()
        if debug:
            print(ask_query)

    def test_topicLinks(self):
        """
        test topic links
        """
        cc = self.getContextContext(wikiId="cr", context_name="CrSchema")
        city_topic = cc.context.topics["City"]
        stl = city_topic.sourceTopicLinks
        ttl = city_topic.targetTopicLinks
        debug = self.debug
        if debug:
            print(json.dumps(stl, indent=2, default=str))
            print(json.dumps(ttl, indent=2, default=str))
        self.assertEqual(1, len(stl))
        self.assertEqual(1, len(ttl))


    def test_Issue8(self):
        """
        https://github.com/WolfgangFahl/pyMetaModel/issues/3
        handle maintopic
        """
        debug = True
        wikiUsers = WikiUser.getWikiUsers(lenient=True)
        if "wiki" in wikiUsers:
            _smwAccess, mw_contexts = self.check_contexts("wiki")
            mw_context = mw_contexts["OpenSourceProjectsContext"]
            context, error, errMsg = Context.fromWikiContext(mw_context, debug=debug)
            self.assertIsNone(error)
            self.assertIsNone(errMsg)
            self.assertTrue("OsProject" in context.topics)
            pass

    def test_Issue23(self):
        """
        https://github.com/WolfgangFahl/pyMetaModel/issues/23
        handle "showInGrid"
        """
        cc=self.getContextContext(context_name="Infrastructure")
        topic=cc.context.topics["Harddisk"]
        debug=self.debug
        #debug=True
        for filterShowInGrid in [False,True]:
            ask_query = topic.askQuery(filterShowInGrid=filterShowInGrid)

            if debug:
                print(f"showInGrid: {filterShowInGrid}")
                print(ask_query)
            self.assertTrue("|?Device name = name" in ask_query)
            if filterShowInGrid:
                self.assertFalse("purpose" in ask_query)


    def test_metamodel_from_wikis(self):
        """
        test reading meta models from wikis
        """
        debug = True
        lenient = True
        for wikiId in WikiUser.getWikiUsers(lenient=True):
            _smwAccess, mw_contexts = self.check_contexts(wikiId)
            for context_name in mw_contexts:
                if debug:
                    print(f"reading context {context_name} from wiki {wikiId}")
                cc = self.getContextContext(wikiId=wikiId, context_name=context_name)
                context = cc.context
                if not lenient:
                    self.assertIsNotNone(context)
                else:
                    if context is None:
                        continue
                if debug:
                    print(f"{context_name} has context {context.name}")
                    print(context)
                    for _name, topic in context.topics.items():
                        print(topic)
                        for _name, property in topic.properties.items():
                            print(property)
                if len(context.errors) > 0:
                    for error in context.errors:
                        print(error)
                if not lenient:
                    self.assertEqual(0, len(context.errors))
                    self.assertEqual(context_name, context.name, wikiId)



"""
Created on 2023-02-25

@author: wf
"""
import os
from meta.metamodel import Context
from tests.basetest import Basetest


class TestSiDIF(Basetest):
    """
    test SiDIF handling
    """
    def setUp(self, debug=False, profile=True):
        Basetest.setUp(self, debug=debug, profile=profile)
        self.examples_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), "examples")


    def check_sidif(self,sidif_path,file)->Context:
        debug=self.debug
        context, error, errMsg = Context.fromSiDIF_input(sidif_path, debug=debug)
        self.assertFalse(error, f"Error in {file}: {errMsg}")
        self.assertIsNone(errMsg, f"Error message in {file}: {errMsg}")
        self.assertIsNotNone(context, f"Context is None for {file}")

        if debug:
            print(f"Processed: {file}")
            print(context)

        # Additional assertions to verify the context
        self.assertIsInstance(context, Context)
        self.assertTrue(len(context.topics) > 0, f"No topics found in {file}")

        # Check if all topics have names
        for topic in context.topics.values():
            self.assertTrue(hasattr(topic, 'name'), f"Topic without name in {file}")

        # Check if all properties have names and types
        for topic in context.topics.values():
            for prop in topic.properties.values():
                self.assertTrue(hasattr(prop, 'name'), f"Property without name in {file}")
                self.assertTrue(hasattr(prop, 'type'), f"Property without type in {file}")
        return context


    def testExamples(self):
        """
        Test all .sidif files in the ../examples directory
        """
        for root, _dirs, files in os.walk(self.examples_dir):
            for file in files:
                if file.endswith(".sidif"):
                    sidif_path = os.path.join(root, file)
                    with self.subTest(file=file):
                        self.check_sidif(sidif_path,file)


    def test_properties_by_index(self):
        """
        test the properties by index handling
        """
        sidif_file="infrastructure/infrastructure.sidif"
        sidif_path = os.path.join(self.examples_dir, sidif_file)
        context = self.check_sidif(sidif_path, sidif_file)
        debug=self.debug
        #debug=True
        for topic_name,topic in context.topics.items():
            props_dict=topic.propertiesDict()
            if debug:
                print(f"Properties Dict for {topic_name}:{topic.documentation}/{topic.wikiDocumentation}")
                for topic_name,(topic_name,props) in enumerate(props_dict.items()):
                    print(f"{topic_name}:")
                    for j,prop in enumerate(props):
                        print(f"  {j:3}:{prop.name} ({getattr(prop,'index','?')})")


    def testInheritance(self):
        """
        test inheritance
        """
        for sidif_file,topics in {
            "infrastructure/infrastructure.sidif": [
                ("Computer","name",["Device"]),
                ("Monitor","name",["Device"]),
                ("Printer","name",["Device"])
            ],
            "scientific-events/scientific-events.sidif":
                [("Event",  "name",["WebWdItem","WikidataItem"]),
                 ("Country","name",["WebWdItem","WikidataItem"]),
                 ("Region" ,"name",["WikidataItem"]),
                 ("Scholar","name",["WebWdItem","WikidataItem"]),
                ],
        }.items():
            sidif_path = os.path.join(self.examples_dir, sidif_file)
            context = self.check_sidif(sidif_path, sidif_file)
            debug=self.debug
            #debug=True
            for topic_name,concept_property_name,extends_list in topics:
                self.assertIn(topic_name, context.topics, f"Topic {topic_name} not found in {sidif_file}")
                topic = context.topics[topic_name]
                self.assertTrue(hasattr(topic, 'extends'), f"Topic {topic_name} in {sidif_file} does not have 'extends' attribute")
                if debug:
                    print(f"{topic_name} extends {topic.extends}")
                extends_topics=topic.get_extends_topics()
                self.assertEqual(len(extends_topics),len(extends_list),topic_name)
                for i,extends_topic in enumerate(extends_topics):
                    expected=extends_list[i]
                    if debug:
                        print(f"{i+1}:{extends_topic.name}={expected}?")
                    self.assertEqual(extends_topic.name,expected)
                props_by_index=topic.propertiesByIndex(to_root=True)
                if debug:
                    for i,prop in enumerate(props_by_index):
                        print(f"{i:3}:{prop.name} ({getattr(prop,'index','?')})")
                self.assertTrue(topic.conceptProperty is not None,topic.name)
                if self.debug:
                    print(f"{topic_name}.{topic.conceptProperty.name} is concept property")
                self.assertEqual(concept_property_name,topic.conceptProperty.name,topic.name)

                pass
            pass


    def testIssue18_readingFromSiDIF_input(self):
        """
        https://github.com/WolfgangFahl/py-sidif/issues/4
        add fromSiDIF_input api to allow reading Context from url or file_path via API
        """
        debug = self.debug
        # debug=True
        for url in [
            "https://raw.githubusercontent.com/WolfgangFahl/pyMetaModel/main/examples/city/city.sidif"
        ]:
            context, error, errMsg = Context.fromSiDIF_input(url, debug=debug)
            self.assertFalse(error)
            self.assertIsNone(errMsg)
            self.assertIsNotNone(context)
            if debug:
                print(context)



"""
Created on 2022-11-23

@author: wf
"""

import sys
import typing
from datetime import datetime
from typing import List, Optional, Tuple

import urllib3
from lodstorage.jsonable import JSONAble
from sidif.sidif import SiDIFParser

from meta.mw import MediaWikiContext


class MetaModelElement(JSONAble):
    """
    a generic MetaModelElement

    to handle the technicalities of being a MetaModelElement so that derived
    MetaModelElements can focus on the MetaModel domain specific aspects
    """

    def __init__(self):
        """
        construct me
        """
        self.__metamodel_props = {}
        cls = self.__class__
        if hasattr(cls, "getSamples"):
            for sample in cls.getSamples():
                for key in sample.keys():
                    if not key in self.__metamodel_props:
                        self.__metamodel_props[key] = key  # Property(self,key)

    def __str__(self):
        """
        get a string representation of me
        """
        text = self.__class__.__name__
        for prop_name in self.__metamodel_props.keys():
            if not isinstance(prop_name, str):
                pass
            elif hasattr(self, prop_name):
                value = getattr(self, prop_name)
                text += f"\n  {prop_name}={str(value)}"
        return text


class Context(MetaModelElement):
    """
    A Context groups some topics like a Namespace/Package.
    This class provides helper functions and constants to render a Context to corresponding wiki pages
    """

    def __init__(self):
        """
        constructor
        """
        MetaModelElement.__init__(self)
        self.topics = {}
        self.types = {}
        self.errors = []

    @classmethod
    def getSamples(cls):
        samples = [
            {
                "name": "MetaModel",
                "since": datetime.strptime("2015-01-23", "%Y-%m-%d"),
                "updated": datetime.strptime("2024-08-07", "%Y-%m-%d"),
                "copyright": "2015-2024 BITPlan GmbH",
                "master": "https://contexts.bitplan.com",
                "demo": "https://wiki.bitplan.com/index.php/List_of_Contexts"
            }
        ]
        return samples

    def sanitize(self,key:str):
        """
        make sure my needed properties exist
        """
        if not hasattr(self,"name"):
            self.name=key
            pass
        pass

    def error(self, msg):
        print(msg, file=sys.stderr)
        self.errors.append(msg)

    def lookupTopic(self, topic_name: str, purpose: str) -> "Topic":
        """
        lookup the given topic_name in my topics for the given purpose

        Args:
            topic_name (str): the name of the topic to lookup
            purpose (str): the purpose to do the lookup for

        Returns:
            Topic: the topic if found else None and an error is added to my errors
        """
        if topic_name in self.topics:
            return self.topics[topic_name]
        else:
            self.error(
                f"""topic {topic_name} not found in context {getattr(self,"name","?")} for {purpose}"""
            )
            return None

    def propertyAlreadyDeclared(
        self, prop_name: str, topic: "Topic", purpose: str
    ) -> bool:
        """
        check whether the given property or role name is already Declared

        Args:
            prop_name (str): the property to check for duplicates
            topic (Topic): the topic to check
            purpose (str): the purpose to be displayed in error messages

        Returns:
            bool: True if this prop_name has already been use
        """
        in_use = prop_name in topic.properties or prop_name in topic.targetTopicLinks
        if in_use:
            self.error(
                f"duplicate name {prop_name} in topic {topic.name} for {purpose}"
            )
        return in_use

    def link_source_with_target(
        self, tl: "TopicLink", source: "Topic", target: "Topic"
    ):
        """
        link the source with the target via the given topicLink

        Args:
            tl (TopicLink): the topicLink
            source (Topic): the source Topic
            target (Topic): the target Topic
        """
        ok = True
        ok = ok and not self.propertyAlreadyDeclared(
            tl.sourceRole, target, f"{tl.name}"
        )
        ok = ok and not self.propertyAlreadyDeclared(
            tl.targetRole, source, f"{tl.name}"
        )
        if ok:
            # n:m handling with two lists on each side
            source.sourceTopicLinks[tl.sourceRole] = tl
            source.targetTopicLinks[tl.targetRole] = tl

    def addLink(self, tl: "TopicLink"):
        """
        link source and target of the given topicLink

        Args:
            tl (TopicLink): the topicLink
        """
        tl.sourceTopic = self.lookupTopic(tl.source, f"topicLink {tl.name}")
        tl.targetTopic = self.lookupTopic(tl.target, f"topicLink {tl.name}")
        if tl.targetTopic is not None and tl.sourceTopic is not None:
            self.link_source_with_target(tl, tl.sourceTopic, tl.targetTopic)

    def addProperty(self, prop: "Property") -> bool:
        """
        add the given property to this context

        Args:
            prop (Property): the property to add

        Returns:
            bool: True if the property adding was successful
        """
        if not hasattr(prop, "topic"):
            self.error(f"prop  {prop} has no topic")
            return False
        topic_name = prop.topic
        if not hasattr(prop, "name"):
            self.error(f"prop {prop} has no name")
            return False
        if not hasattr(prop, "type"):
            prop.type = "Text"
        topic = self.lookupTopic(topic_name, f"property {prop.name}")
        if topic:
            topic.properties[prop.name] = prop
            return True

    def createProperty4TopicLink(self, tl: "TopicLink") -> "Property":
        """
        create a property for the given topic link

        Args:
            tl (TopicLink):  the topiclink to create a property for
        """
        # see https://wiki.bitplan.com/index.php/SiDIFTemplates#properties
        prop = Property()
        prop.name = tl.sourceRole
        prop.label = prop.name
        if hasattr(tl, "sourceDocumentation"):
            prop.documentation = tl.sourceDocumentation
        else:
            prop.documentation = f"{prop.name}"
        prop.topic = tl.target
        prop.type = "Page"
        prop.topicLink = tl
        prop.isLink = True
        return prop

    @classmethod
    def fromDictOfDicts(cls, did: dict) -> "Context":
        """
        fill me from the given dict of dicts

        Args:
            did (dict): the dict of dicts

        Returns:
            Context: the context read
        """
        context = None
        for key, record in did.items():
            isA = record.get("isA")
            if isA == "Context":
                context = Context()
                context.fromDict(record)
                context.sanitize(key)
            elif isA == "TopicLink":
                """
                # Event n : 1 City
                Event_in_City isA TopicLink
                "eventInCity" is name of it
                "city" is sourceRole of it
                false is sourceMultiple of it
                "City" is source of it
                "event" is targetRole of it
                true is targetMultiple of it
                "Event" is target of it
                """
                tl = TopicLink()
                tl.fromDict(record)
                context.addLink(tl)
            elif isA == "Property":
                prop = Property()
                prop.isLink = False
                prop.fromDict(record)
                prop.sanitize()
                context.addProperty(prop)
            else:  # isA == Topic or in declared topics
                topic = Topic()
                topic.fromDict(record)
                topic.sanitize()
                if context is None:
                    context = Context()
                    context.name = "GlobalContext"
                    context.since = "2022-11-26"
                    context.master = "http://contexts.bitplan.com"
                    context.error(f"topic {topic.name} has no defined context")
                if hasattr(topic, "name"):
                    context.topics[topic.name] = topic
                    topic.context_obj=context
                elif hasattr(topic, "type"):
                    context.types[topic.type]=topic
                else:
                    # potential foreign or extends declaration
                    context.error(f"missing name for topic {topic} {key} - foreign declaration?")

        # link topic to concepts and add topicLinks
        for topic in context.topics.values():
            topic.setConceptProperty()
            for _tl_name, tl in topic.targetTopicLinks.items():
                prop = context.createProperty4TopicLink(tl)
                context.addProperty(prop)
                pass
        return context

    @classmethod
    def fromSiDIF_input(cls, sidif_input: str, debug: bool = False)->Tuple['Context',str,str]:
        """
        initialize me from the given SiDIF input which might be a file path
        or url

        Args:
            sidif_input (str): path to local file or URL
            debug (bool): if True swith debugging on

        Returns:
            Tuple[Context,str,str]: context, error and errorMessage
        """
        if sidif_input.startswith("http:") or sidif_input.startswith("https:"):
            url = sidif_input
            http = urllib3.PoolManager()
            response = http.request("GET", url)
            sidif = response.data.decode("utf-8")
        else:
            sidif_path = sidif_input
            with open(sidif_path, "r") as sidif_file:
                sidif = sidif_file.read()
        return Context.fromSiDIF(sidif, title=sidif_input, debug=debug)

    @classmethod
    def fromSiDIF(
        cls, sidif: str, title: str, depth: int = None, debug: bool = False
    ) -> typing.Tuple["Context", object, str]:
        """
        initialize me from the given SiDIF markup

        Args:
            sidif (str): the SiDIF markup to parse
            title (str): the title for the SiDIF
            depth (int): the explain depth to show for the errorMessage
            debug (bool): if True handle debugging

        Returns:
            Tuple[Context,str,str]: context, error and errorMessage

        """
        errMsg = None
        context = None
        sp = SiDIFParser(debug=debug)
        parsed, error = sp.parseText(sidif, title, depth=depth)
        if debug:
            sp.printResult(parsed)
        if error is None:
            dif = parsed[0]
            did = dif.toDictOfDicts()
            context = Context.fromDictOfDicts(did)
            context.dif = dif
            context.did = did
        else:
            errMsg = sp.errorMessage(title, error, depth=depth)
        return context, error, errMsg

    @classmethod
    def fromWikiContext(
        cls, mw_context: MediaWikiContext, depth: int = None, debug: bool = False
    ) -> Tuple["Context", Optional[Exception], Optional[str]]:
        """
        initialize me from the given MediaWiki Context

        Args:
            mw_context (MediaWikiContext): the Mediawiki context
            depth (int): the explain depth to show for the errorMessage
            debug (bool): if True handle debugging

        Return:
            tuple(Context,Exception,str): the metamodel and potential parsing errors as Exception and error Message
        """
        context = None
        error = None
        errMsg = None
        sidif = None
        if debug:
            print(f"reading sidif for {mw_context.context} from {mw_context.wikiId}")
        try:
            sidif = mw_context.read_sidif()
        except BaseException as ex:
            error = ex
            errMsg = str(ex)
        if sidif is not None:
            context, error, errMsg = cls.fromSiDIF(
                sidif=sidif, title=mw_context.wikiId, depth=depth, debug=debug
            )
        return context, error, errMsg


class Topic(MetaModelElement):
    """
    A Topic is a Concept/Class/Thing/Entity
    """

    def __init__(self):
        """
        constructor
        """
        MetaModelElement.__init__(self)
        self._pluralName = None # Initialize with underscore to indicate a protected attribute

        self.properties = {}
        self.sourceTopicLinks = {}
        self.targetTopicLinks = {}

    def get_extends_topics(self, l: List['Topic'] = None) -> List['Topic']:
        """
        Get the topics this topic is extending.

        Args:
        l (List['Topic']): an optional list to extend. If None, a new list is created.

        Returns:
        List['Topic']: A list of topics that this topic extends.
        """
        if l is None:
            l = []

        extends = getattr(self, "extends", None)

        # recursive inheritance
        if extends:
            extends_topic = self.context_obj.lookupTopic(extends, purpose=f"extends")
            if extends_topic and extends_topic not in l:
                l.append(extends_topic)
                extends_topic.get_extends_topics(l)

        return l

    def get_all_properties(self) -> List['Property']:
        """
        Get all properties of the topic, including those inherited from extended topics.

        Returns:
            List[Property]: A list of all properties, including inherited ones.
        """
        all_properties = list(self.properties.values())

        # Get extended topics
        extended_topics = self.get_extends_topics()

        # Add properties from extended topics
        for extended_topic in extended_topics:
            for prop in extended_topic.properties.values():
                if prop.name not in [p.name for p in all_properties]:
                    all_properties.append(prop)

        return all_properties


    def sanitize(self):
        """
        make sure my properties exist note that name is not handled here
        on purpose since it needs special treatment
        """
        doc=self.documentation if hasattr(self, "documentation") else "?"
        if not hasattr(self, "wikiDocumentation") or not self.wikiDocumentation:
            self.wikiDocumentation=doc
        if not hasattr(self,"defaultstoremode"):
            self.defaultstoremode="property"

    @classmethod
    def getSamples(cls):
        samples = [
            {
                "pageTitle": "Concept:Topic",
                "name": "Topic",
                "pluralName": "Topics",
                "icon": "File:Topic_icon.png",
                "iconUrl": "/images/a/ae/Index.png",
                "documentation": "A Topic is a Concept/Class/Thing/Entity",
                "wikiDocumentation": "A Topic is a Concept/Class/Thing/Entity",
                "context": "MetaModel",
                "listLimit": 7,
                "cargo": True,
            }
        ]
        return samples

    def setConceptProperty(self):
        """
        set my concept property to any primary key or mandatory property
        """
        self.conceptProperty = None
        all_properties=self.propertiesByIndex(to_root=True)
        for prop in all_properties:
            mandatory = False
            primaryKey = False
            if hasattr(prop, "mandatory"):
                mandatory = prop.mandatory
            if hasattr(prop, "primaryKey"):
                primaryKey = prop.primaryKey
            if mandatory or primaryKey:
                self.conceptProperty = prop
                break

    @property
    def pluralName(self)->str:
        """
        Getter for pluralName.

        Returns:
            str: The plural name of the topic.
        """
        # Default behavior if _pluralName is not explicitly set
        return self._pluralName if self._pluralName is not None else f"{self.name}s"

    @pluralName.setter
    def pluralName(self, value):
        """
        Setter for pluralName.

        Args:
            value (str): The plural name to be set for the topic.
        """
        self._pluralName = value

    def getPluralName(self) -> str:
        """
        get the plural name for this topic

        Returns:
            str: the pluralname e.g. "Topics" for "Topic" or "Status" for "Status" or
            "Entities" for "Entity"
        """
        return self.pluralName

    def getListLimit(self) -> int:
        """
        get the list limit for this topic
        """
        listLimit = getattr(self, "listLimit", 200)
        return listLimit

    def sortProperties(self) -> list:
        """
        get the properties that we should sort by

        Returns:
            list: a list of properties in sort order
        """
        prop_list = []
        all_properties=self.get_all_properties()
        for prop in all_properties:
            if hasattr(prop, "sortPos"):
                sortPos = prop.sortPos
                if sortPos:
                    prop_list.append(prop)
        prop_list = sorted(prop_list, key=lambda prop: prop.sortPos)
        return prop_list

    def propertiesDict(self) -> dict:
        """
        Return a dictionary mapping topic names to lists of properties,
        considering the inheritance chain and sorting properties locally by index.

        Returns:
            dict: A dictionary with topic names as keys and sorted lists of properties as values.
        """
        properties_dict = {}
        inheritance_chain = self.get_extends_topics()
        inheritance_chain.append(self)
        for topic in inheritance_chain:
            sorted_props=topic.propertiesByIndex(to_root=False)
            properties_dict[topic.name] = sorted_props

        return properties_dict


    def propertiesByIndex(self,to_root:bool=False) -> list:
        """
        Return the properties by index, considering inheritance chain.

        Args:
            to_root(bool): if True follow the inheritance chain to the root

        Returns:
            list: The list of properties sorted by index,
            with inheritance considered.
        """
        if to_root:
            sorted_props = []
            for _topic_name, props in self.propertiesDict().items():
                sorted_props.extend(props)
        else:
            props = list(self.properties.values())
            props_count = len(props)

            # Sort properties locally by index
            sorted_props = sorted(props, key=lambda p: getattr(p, 'index', props_count))

        return sorted_props


    def askSort(self) -> str:
        """
        generate the sort clause for my SMW ask query

        Returns:
            str: the generated wiki markup
        """
        sort = ""
        order = ""
        delim = ""
        sortproperties = self.sortProperties()
        for prop in sortproperties:
            direction = (
                "ascending" if getattr(prop, "sortAscending", True) else "descending"
            )
            sort += delim + f"{self.name} {prop.name}"
            order += delim + direction
            delim = ","
            pass
        sortClause = f"|sort={sort}\n" if sort else ""
        orderClause = f"|order={order}\n" if order else ""
        markup = f"{sortClause}{orderClause}"
        return markup

    def askQuery(
        self,
        mainlabel: str = None,
        filterShowInGrid: bool = True,
        listLimit: int = None,
    ) -> str:
        """
        Get the askQuery for the topic, considering inheritance.

        Args:
            mainlabel (str): the mainlabel to use - topic.name as default
            filterShowInGrid (bool): if True include only properties with showInGrid not being false
            listLimit (int): the list limit to use
        Returns:
            str: the markup for the query
        """
        if listLimit is None:
            listLimit = self.getListLimit()
        if mainlabel is None:
            mainlabel = self.name
        markup = f"""{{{{#ask: [[Concept:{self.name}]]
|mainlabel={mainlabel}
"""
        properties_dict = self.propertiesDict()
        for topic_name, properties in properties_dict.items():
            for prop in properties:
                if filterShowInGrid:
                    show=getattr(prop, "showInGrid",False)
                else:
                    show = True
                if show:
                    # Use the topic_name as prefix, which will be correct for inherited properties
                    markup += f"|?{topic_name} {prop.name} = {prop.name}\n"
        markup += f"|limit={listLimit}\n"
        markup += f"""{self.askSort()}}}}}"""
        return markup

class Property(MetaModelElement):
    """
    Provides helper functions and constants for properties
    """

    def __init__(self):
        """
        constructor
        """
        MetaModelElement.__init__(self)

    @classmethod
    def getSamples(cls):
        samples = [
            {
                "name": "Title",
                "label": "Offical Name",
                "type": "Special:Types/Text",
                "index": 2,
                "sortPos": 2,
                "primaryKey": False,
                "mandatory": True,
                "namespace": "Main",
                "scope": "Global",
                "size": 25,
                "uploadable": False,
                "defaultValue": "Some Title",
                "inputType": "combobox",
                "allowedVaues": "Tile A, Title B",
                "documentation": " Documentation can contain\n line breaks",
                "values_from": "property=Title",
                "showInGrid": False,
                "isLink": False,
                "nullable": False,
                "topic": "Concept:Event",
            },
            {
                "name": "wikidataid",
                "type": "Special:Types/External identifier",
                "formatterURI": "https://www.wikidata.org/wiki/$1",
            },
            # Properties that are not included in the MetaModel
            {
                "placeholder": "e.g. SMWCon",
                "regexp": "NaturalNumber",
                "usedFor": "Concept:Event, Concept:Event series",
                "pageTitle": "Property:Title",
            },
        ]
        return samples


    def sanitize(self):
        """
        make sure attributes such as namespace are set properly
        """
        self.namespace="Property"
        if hasattr(self, "scope"):
            pass
        pass

class TopicLink(MetaModelElement):
    """
    A TopicLink links two Concepts/Topics
    """

    @classmethod
    def getSamples(cls):
        samples = [
            {
                "isA": "TopicLink",
                "name": "containedProperties",
                "sourceRole": "topic",
                "sourceMultiple": False,
                "source": "Topic",
                "targetRole": "properties",
                "targetMultiple": True,
                "target": "Property",
            },
            {
                "isA": "TopicLink",
                "name": "containedTopics",
                "sourceRole": "context",
                "sourceMultiple": False,
                "source": "Context",
                "targetRole": "topics",
                "targetMultiple": True,
                "target": "Topic",
            },
            {
                "isA": "TopicLink",
                "name": "typeOfProperty",
                "sourceRole": "usedByProperties",
                "sourceMultiple": True,
                "source": "Property",
                "sourceDocumentation": "the properties having this type",
                "targetRole": "smw_type",
                "targetMultiple": False,
                "target": "SMW_Type",
                "masterDetail": False,
                "targetDocumentation": "the smw_type being used by this property",
            },
        ]
        return samples



"""
Created on 2023-02-20

@author: wf
"""

from meta.metamodel_cmd import MetaModelCmd
from meta.sidif2linkml import SiDIF2LinkML
from tests.basemwtest import BaseMediawikiTest


class TestLinkML(BaseMediawikiTest):
    """
    test linkml handling

    https://linkml.io/
    """

    def setUp(self, debug=False, profile=True):
        """
        setUp
        """
        BaseMediawikiTest.setUp(self, debug=debug, profile=profile)
        for wikiId in ["wiki", "cr"]:
            self.getWikiUser(wikiId, save=True)

    def testSiDIF2LinkML(self):
        """
        test conversion from SiDIF to LinkML
        """
        debug = self.debug
        debug = True
        parser = MetaModelCmd.getArgParser()
        mm_cmd = MetaModelCmd(debug=debug)
        for wikiId, context_name in [("wiki", "MetaModel"), ("cr", "CrSchema")]:
            args = parser.parse_args(["--wikiId", wikiId, "--context", context_name])
            mm_cmd.readContext(args)
            self.assertTrue(not mm_cmd.hasError())
            s2l = SiDIF2LinkML(mm_cmd.context)
            yaml_text = s2l.asYaml()
            if debug:
                print(yaml_text)



"""
Created on 2023-02-20

@author: wf
"""

from linkml.generators.linkmlgen import LinkmlGenerator
from linkml.utils.schema_builder import SchemaBuilder
from linkml_runtime.linkml_model import (
    ClassDefinition,
    SlotDefinition,
)
from linkml_runtime.utils.schemaview import SchemaView

from meta.metamodel import Context


class SiDIF2LinkML:
    """
    converter between SiDIF and LINKML
    """

    def __init__(self, context: Context):
        self.context = context

    def asYaml(self, common_property_slots: bool = True, delim="_") -> str:
        """
        convert my context

        Args:
            common_property_slots (bool): if True reuse slots

        Returns:
            str: the yaml markup

        """
        context = self.context
        sb = SchemaBuilder(id=context.name, name=context.name)
        # https://linkml.io/linkml-model/docs/SchemaDefinition/
        sb.add_defaults()
        sd = sb.schema
        sv = SchemaView(sd)
        if hasattr(context, "copyright"):
            copyright_str = f" copyright {context.copyright}"
        else:
            copyright_str = ""
        if hasattr(context, "master"):
            master = context.master
        else:
            master = "http://example.com"
        uri = f"{master}/{context.name}"
        for topic in self.context.topics.values():
            cd = ClassDefinition(name=topic.name)
            cd.description = topic.documentation
            sv.add_class(cd)
            for prop in topic.properties.values():
                slot = None
                if common_property_slots:
                    qname = prop.name
                    if prop.name in sd.slots:
                        slot = sd.slots[prop.name]
                        slot.description += "," + prop.documentation
                else:
                    qname = f"{topic.name}{delim}{prop.name}"
                if slot is None:
                    slot = SlotDefinition(name=qname)
                    if hasattr(prop, "documentation"):
                        slot.description = prop.documentation
                    typesMap = {
                        "Boolean": "boolean",
                        "Code": "string",  # @TODO will need special type
                        "Date": "date",
                        "Text": "string",
                        "URL": "uri",
                        "Page": "string",  # @TODO will need own type
                        "Number": "double",  # @TODO will need special type handling
                    }
                    if prop.isLink:
                        # 1:1 and 1:n handling
                        slot.range = prop.topicLink.source
                    else:
                        if prop.type in typesMap:
                            slot.range = typesMap[prop.type]
                        else:
                            slot.range = "string"
                    sv.add_slot(slot)
                cd.attributes[qname] = slot
                pass

            pass

        lml_gen = LinkmlGenerator(schema=sd, format="yaml")
        yaml_text = lml_gen.serialize()
        return yaml_text
